/*
 *  Velocity MIB - system group implementation - velocity_system.c
 *
 */
 
#include <net-snmp/net-snmp-config.h>
#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
 
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
 
#include "host.h"
#include "host_res.h"
#include "velocity_system.h"
#include <net-snmp/agent/auto_nlist.h>
 
#ifdef HAVE_SYS_PROC_H
#include <sys/param.h>
#include "sys/proc.h"
#endif
#if HAVE_UTMPX_H
#include <utmpx.h>
#else
#include <utmp.h>
#endif
 
#ifdef linux
#ifdef HAVE_LINUX_TASKS_H
#include <linux/tasks.h>
#else
/*
 * If this file doesn't exist, then there is no hard limit on the number
 * of processes, so return 0 for velocitySystemMaxProcesses.
 */
#define NR_TASKS	0
#endif
#endif
 
#if defined(hpux10) || defined(hpux11)
#include <sys/pstat.h>
#endif
 
#ifdef HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif
 
#if !defined(UTMP_FILE) && defined(_PATH_UTMP)
#define UTMP_FILE _PATH_UTMP
#endif
 
#if defined(UTMP_FILE) && !HAVE_UTMPX_H
void            setutent(void);
void            endutent(void);
struct utmp    *getutent(void);
#endif                          /* UTMP_FILE */
 
 
        /*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/
 
static int      get_load_dev(void);
static int      count_users(void);
extern int      count_processes(void);
 
 
        /*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/
 
#define	VELOCITYSYS_UPTIME		1
#define	VELOCITYSYS_DATE		2
#define	VELOCITYSYS_LOAD_DEV		3
#define	VELOCITYSYS_LOAD_PARAM	4
#define	VELOCITYSYS_USERS		5
#define	VELOCITYSYS_PROCS		6
#define	VELOCITYSYS_MAXPROCS		7
#define	VELOCITYSYS_PAGEIN              8
#define	VELOCITYSYS_PAGEOUT             9
#define	VELOCITYSYS_SWAPIN             10
#define	VELOCITYSYS_SWAPOUT            11
#define	VELOCITYSYS_INTR               12
#define	VELOCITYSYS_CTXT               13
#define	VELOCITYSYS_PROCTOTAL          14
 
#define	VELOCITY_TYPE_PAGE 101
#define	VELOCITY_TYPE_INTR 102
#define	VELOCITY_TYPE_CTXT 103
#define	VELOCITY_TYPE_SWAP 104
#define	VELOCITY_TYPE_PROCTOTAL 105
 
#define VELOCITY_IN 200
#define VELOCITY_OUT 201
 
struct variable2 velocitysystem_variables[] = {
    {VELOCITYSYS_UPTIME, ASN_TIMETICKS, RONLY, var_velocitysys, 1, {1}},
    {VELOCITYSYS_DATE, ASN_OCTET_STR, RONLY, var_velocitysys, 1, {2}},
    {VELOCITYSYS_LOAD_DEV, ASN_INTEGER, RONLY, var_velocitysys, 1, {3}},
    {VELOCITYSYS_LOAD_PARAM, ASN_OCTET_STR, RONLY, var_velocitysys, 1, {4}},
    {VELOCITYSYS_USERS, ASN_GAUGE, RONLY, var_velocitysys, 1, {5}},
    {VELOCITYSYS_PROCS, ASN_GAUGE, RONLY, var_velocitysys, 1, {6}},
    {VELOCITYSYS_MAXPROCS, ASN_INTEGER, RONLY, var_velocitysys, 1, {7}},
    {VELOCITYSYS_PAGEIN, ASN_INTEGER, RONLY, var_velocitysys, 1, {8}},
    {VELOCITYSYS_PAGEOUT, ASN_INTEGER, RONLY, var_velocitysys, 1, {9}},
    {VELOCITYSYS_SWAPIN, ASN_INTEGER, RONLY, var_velocitysys, 1, {10}},
    {VELOCITYSYS_SWAPOUT, ASN_INTEGER, RONLY, var_velocitysys, 1, {11}},
    {VELOCITYSYS_INTR, ASN_INTEGER, RONLY, var_velocitysys, 1, {12}},
    {VELOCITYSYS_CTXT, ASN_INTEGER, RONLY, var_velocitysys, 1, {13}},
    {VELOCITYSYS_PROCTOTAL, ASN_INTEGER, RONLY, var_velocitysys, 1, {14}}
};
oid             velocitysystem_variables_oid[] = { 1, 3, 6, 1, 4, 1, 15601, 1 };
 
#ifdef linux
int             linux_stat(int, int);
#endif
 
 
 
void
init_velocity_system(void)
{
#ifdef NPROC_SYMBOL
    auto_nlist(NPROC_SYMBOL, 0, 0);
#endif
 
    REGISTER_MIB("velocity/velocity_system", velocitysystem_variables, variable2,
                 velocitysystem_variables_oid);
}
 
/*
 * header_velocitysys(...
 * Arguments:
 * vp     IN      - pointer to variable entry that points here
 * name    IN/OUT  - IN/name requested, OUT/name found
 * length  IN/OUT  - length of IN/OUT oid's
 * exact   IN      - TRUE if an exact match was requested
 * var_len OUT     - length of variable or 0 if function returned
 * write_method
 */
 
int
header_velocitysys(struct variable *vp,
             oid * name,
             size_t * length,
             int exact, size_t * var_len, WriteMethod ** write_method)
{
#define VELOCITYSYS_NAME_LENGTH	9
    oid             newname[MAX_OID_LEN];
    int             result;
 
    DEBUGMSGTL(("velocity/velocity_system", "var_velocitysys: "));
    DEBUGMSGOID(("velocity/velocity_system", name, *length));
    DEBUGMSG(("velocity/velocity_system", " %d\n", exact));
 
    memcpy((char *) newname, (char *) vp->name, vp->namelen * sizeof(oid));
    newname[VELOCITYSYS_NAME_LENGTH] = 0;
    result = snmp_oid_compare(name, *length, newname, vp->namelen + 1);
    if ((exact && (result != 0)) || (!exact && (result >= 0)))
        return (MATCH_FAILED);
    memcpy((char *) name, (char *) newname,
           (vp->namelen + 1) * sizeof(oid));
    *length = vp->namelen + 1;
 
    *write_method = 0;
    *var_len = sizeof(long);    /* default to 'long' results */
    return (MATCH_SUCCEEDED);
}
 
int
linux_stat(int stat_type, int in_or_out)
{
    FILE           *fp;
    char            buf[100];
    int             in = -1, out = -1;
 
    if ((fp = fopen("/proc/stat", "r")) == NULL)
        return -1;
 
    while (fgets(buf, sizeof(buf), fp) != NULL) {
        if ((!strncmp(buf, "page", 4) && stat_type == VELOCITY_TYPE_PAGE) ||
            (!strncmp(buf, "intr", 4) && stat_type == VELOCITY_TYPE_INTR) ||
	    (!strncmp(buf, "ctxt", 4) && stat_type == VELOCITY_TYPE_CTXT) ||
	    (!strncmp(buf, "processes", 9) && stat_type == VELOCITY_TYPE_PROCTOTAL) ||
	    (!strncmp(buf, "swap", 4) && stat_type == VELOCITY_TYPE_SWAP)) {
            sscanf(buf, "%*s %d %d", &in, &out);
	    DEBUGMSGTL(("string found", " %s  type - %d\n", buf, stat_type));
            break;
        }
	else {
	    DEBUGMSGTL(("string not found", " %s type - %d\n", buf, stat_type ));
	}
    }
 
    fclose(fp);
    return (in_or_out == VELOCITY_IN ? in : out);
 
}
 
 
 
        /*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/
 
u_char         *
var_velocitysys(struct variable * vp,
          oid * name,
          size_t * length,
          int exact, size_t * var_len, WriteMethod ** write_method)
{
    static char     string[100];
    time_t          now;
#ifndef NR_TASKS
    int             nproc = 0;
#endif
#ifdef linux
    FILE           *fp;
#endif
#if CAN_USE_SYSCTL && defined(CTL_KERN) && defined(KERN_MAXPROC)
    static int      maxproc_mib[] = { CTL_KERN, KERN_MAXPROC };
    int             buf_size;
#endif
#if defined(hpux10) || defined(hpux11)
    struct pst_static pst_buf;
#endif
 
    if (header_velocitysys(vp, name, length, exact, var_len, write_method) ==
        MATCH_FAILED)
        return NULL;
 
    switch (vp->magic) {
    case VELOCITYSYS_UPTIME:
        long_return = get_uptime();
        return (u_char *) & long_return;
    case VELOCITYSYS_DATE:
        (void *) time(&now);
        return (u_char *) date_n_time(&now, var_len);
    case VELOCITYSYS_LOAD_DEV:
        long_return = get_load_dev();
        return (u_char *) & long_return;
    case VELOCITYSYS_LOAD_PARAM:
#ifdef linux
        fp = fopen("/proc/cmdline", "r");
        fgets(string, sizeof(string), fp);
        fclose(fp);
#else
#if NO_DUMMY_VALUES
        return NULL;
#endif
        sprintf(string, "ask Dave");    /* XXX */
#endif
        *var_len = strlen(string);
        return (u_char *) string;
    case VELOCITYSYS_PAGEIN:
        long_return = linux_stat(VELOCITY_TYPE_PAGE,VELOCITY_IN);
        return (u_char *) & long_return;
    case VELOCITYSYS_PAGEOUT:
        long_return = linux_stat(VELOCITY_TYPE_PAGE,VELOCITY_OUT);
        return (u_char *) & long_return;
    case VELOCITYSYS_SWAPIN:
        long_return = linux_stat(VELOCITY_TYPE_SWAP,VELOCITY_IN);
        return (u_char *) & long_return;
    case VELOCITYSYS_SWAPOUT:
        long_return = linux_stat(VELOCITY_TYPE_SWAP,VELOCITY_OUT);
        return (u_char *) & long_return;
    case VELOCITYSYS_CTXT:
        long_return = linux_stat(VELOCITY_TYPE_CTXT,VELOCITY_IN);
        return (u_char *) & long_return;
    case VELOCITYSYS_PROCTOTAL:
        long_return = linux_stat(VELOCITY_TYPE_PROCTOTAL,VELOCITY_IN);
        return (u_char *) & long_return;
    case VELOCITYSYS_INTR:
#ifdef linux
        long_return = linux_stat(VELOCITY_TYPE_INTR,VELOCITY_IN);
#endif
        return (u_char *) & long_return;
    case VELOCITYSYS_USERS:
        long_return = count_users();
        return (u_char *) & long_return;
    case VELOCITYSYS_PROCS:
#if USING_VELOCITY_VELOCITY_SWRUN_MODULE
        long_return = count_processes();
#else
#if NO_DUMMY_VALUES
        return NULL;
#endif
        long_return = 0;
#endif
        return (u_char *) & long_return;
    case VELOCITYSYS_MAXPROCS:
#if defined(NR_TASKS)
        long_return = NR_TASKS; /* <linux/tasks.h> */
#elif CAN_USE_SYSCTL && defined(CTL_KERN) && defined(KERN_MAXPROC)
        buf_size = sizeof(nproc);
        if (sysctl(maxproc_mib, 2, &nproc, &buf_size, NULL, 0) < 0)
            return NULL;
        long_return = nproc;
#elif defined(hpux10) || defined(hpux11)
        pstat_getstatic(&pst_buf, sizeof(struct pst_static), 1, 0);
        long_return = pst_buf.max_proc;
#elif defined(NPROC_SYMBOL)
        auto_nlist(NPROC_SYMBOL, (char *) &nproc, sizeof(int));
        long_return = nproc;
#else
#if NO_DUMMY_VALUES
        return NULL;
#endif
        long_return = 0;
#endif
        return (u_char *) & long_return;
    default:
        DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_velocitysys\n",
                    vp->magic));
    }
    return NULL;
}
 
 
        /*********************
	 *
	 *  Internal implementation functions
	 *
	 *********************/
 
                /*
                 *  Return the DeviceIndex corresponding
                 *   to the boot device
                 */
static int
get_load_dev(void)
{
    return (HRDEV_DISK << HRDEV_TYPE_SHIFT);    /* XXX */
}
 
static int
count_users(void)
{
    int             total = 0;
#if HAVE_UTMPX_H
#define setutent setutxent
#define getutent getutxent
#define endutent endutxent
    struct utmpx   *utmp_p;
#else
    struct utmp    *utmp_p;
#endif
 
    setutent();
    while ((utmp_p = getutent()) != NULL) {
#ifndef UTMP_HAS_NO_TYPE
        if (utmp_p->ut_type == USER_PROCESS)
#endif
            ++total;
    }
    endutent();
    return total;
}
 
#if defined(UTMP_FILE) && !HAVE_UTMPX_H
 
static FILE    *utmp_file;
static struct utmp utmp_rec;
 
void
setutent(void)
{
    if (utmp_file)
        fclose(utmp_file);
    utmp_file = fopen(UTMP_FILE, "r");
}
 
void
endutent(void)
{
    if (utmp_file) {
        fclose(utmp_file);
        utmp_file = NULL;
    }
}
 
struct utmp    *
getutent(void)
{
    if (!utmp_file)
        return NULL;
    while (fread(&utmp_rec, sizeof(utmp_rec), 1, utmp_file) == 1)
        if (*utmp_rec.ut_name && *utmp_rec.ut_line)
            return &utmp_rec;
    return NULL;
}
 
#endif                          /* UTMP_FILE */
